{ File   : FSB2WAV.dpr
  Author : Deniz Oezmen
  Created: 2008-12-08
  Changed: 2010-08-29

  A Vampyre Story (FMOD) FSB audio converter.
}
program FSB2WAV;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  SysUtils,
  StrUtils,
  IMAHeader in 'backend\IMAHeader.pas',
  Utils     in 'utils\Utils.pas';

const
  FSBHeaderID = 'FSB4';
  FEVHeaderID = 'FEV1';

type
  TFSBHeader = packed record
    ID        : array[1..4] of Char;
    NumEntries: Longword;
    DirSize   : Longword;
    DataSize  : Longword;
    Unknown1  : Longword;
    Unknown2  : array[0..11] of Byte;
    Unknown3  : Single;
    Unknown4  : Single;
    Unknown5  : Single;
    Unknown6  : Single;
  end;

  TFSBEntry = packed record
    EntrySize  : Word;
    Name       : array[1..30] of Char;
    Unknown1   : Longword;
    DataSize   : Longword;
    Unknown2   : Longword;
    Unknown3   : Longword;
    BlockSize  : Byte;
    Unknown4   : array[1..3] of Byte;
    Frequency  : Longword;
    Unknown5   : Word;
    Unknown6   : Word;
    Unknown7   : Word;
    NumChannels: Word;
    Unknown8   : Single;
    Unknown9   : Single;
    Unknown10  : Longword;
    Unknown11  : Longword;
  end;

  TFEVHeader = packed record
    ID        : array[1..4] of Char;
    Unknown   : Longword;
    NameLen   : Longword;
    Name      : AnsiString;
    NumEntries: Longword;
  end;

  TFEVEntry = packed record
    Unknown1: Longword;
    Unknown2: Longword;
    NameLen : Longword;
    Name    : string;
  end;

var
  WAVHeader: packed array[0..43] of Char =
    ( 'R',  'I',  'F',  'F', #$00, #$00, #$00, #$00,  'W',  'A',  'V',  'E',
      'f',  'm',  't',  ' ', #$10, #$00, #$00, #$00, #$01, #$00, #$01, #$00,
     #$22, #$56, #$00, #$00, #$22, #$56, #$00, #$00, #$02, #$00, #$10, #$00,
     'd',  'a',  't',  'a',  #$00, #$00, #$00, #$00);
  FEVBuffer: array of Byte;
  FSBList  : array of AnsiString;

// pretty "hack-y" procedure to get naming references without really
// understanding the FEV file structure
procedure LoadFEVBuffer(const FEVFileName: AnsiString);
var
  i        : Longint;
  FEVFile  : file;
  FEVHeader: TFEVHeader;
  FEVEntry : TFEVEntry;
begin
  SetLength(FEVBuffer, 0);

  if FEVFileName = '' then
    Exit;

  FileMode := fmOpenRead;
  AssignFile(FEVFile, FEVFileName);

  {$I-}
  Reset(FEVFile, 1);
  {$I-}
  if IOResult <> 0 then
  begin
    WriteLn(Format('WARNING: Could not open FEV file ''%s'' for naming refere' +
      'nce.', [FEVFileName]));
    Exit
  end;

  {$I-}
  BlockRead(FEVFile, FEVHeader, 12);
  IF (IOResult <> 0) or (FEVHeader.ID <> FEVHeaderID) then
  begin
    WriteLn('WARNING: Not a valid FEV file.');
    CloseFile(FEVFile);
    Exit
  end;
  SetLength(FEVHeader.Name, FEVHeader.NameLen);
  BlockRead(FEVFile, FEVHeader.Name[1], FEVHeader.NameLen);
  BlocKRead(FEVFile, FEVHeader.NumEntries, 4);

  SetLength(FSBList, FEVHeader.NumEntries);

  for i := 0 to High(FSBList) do
  begin
    BlockRead(FEVFile, FEVEntry, 12);
    SetLength(FEVEntry.Name, FEVEntry.NameLen);
    BlockRead(FEVFile, FEVEntry.Name[1], FEVEntry.NameLen);
    FSBList[i] := PChar(@FEVEntry.Name[1])
  end;
  {$I+}

  if IOResult <> 0 then
  begin
    WriteLn('WARNING: Could not read from FEV file.');
    CloseFile(FEVFile);
    Exit
  end;

  SetLength(FEVBuffer, FileSize(FEVFile) - FilePos(FEVFile));
  FillChar(FEVBuffer[0], Length(FEVBuffer), 0);
  {$I-}
  BlockRead(FEVFile, FEVBuffer[0], Length(FEVBuffer));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('WARNING: Could not read from FEV file.');
    CloseFile(FEVFile);
    Exit
  end;

  CloseFile(FEVFile)
end;

function OpenFSBFile(var FSBFile: file; const FSBName: AnsiString): Boolean;
begin
  FileMode := fmOpenRead;
  AssignFile(FSBFile, FSBName);

  {$I-}
  Reset(FSBFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenWAVFile(var WAVFile: file; WAVName: AnsiString): Boolean;

  function MakeDir(const FileName: string): Boolean;
  var
    Dir,
    SubDir: string;
    p     : Longint;
  begin
    Dir := ExtractFileDir(FileName);

    if Dir = '' then
    begin
      Result := True;
      Exit
    end;

    p := Pos('\', Dir);
    while p > 0 do
    begin
      SubDir := Copy(Dir, 1, p - 1);
      if not DirectoryExists(SubDir) then
      begin
        Result := CreateDir(SubDir);
        if not Result then
          Exit
      end;
      p := PosEx('\', Dir, p + 1)
    end;

    if not DirectoryExists(Dir) then
      Result := CreateDir(Dir)
    else
      Result := True
  end;

begin
  WAVName := AnsiReplaceStr(WAVName, '/', '\');
  if AnsiLowerCase(ExtractFileExt(WAVName)) <> '.wav' then
    WAVName := ChangeFileExt(WAVName, '.wav');

  WriteLn('  ', WAVName);

  Result := MakeDir(WAVName);
  if not Result then
    Exit;

  FileMode := fmOpenWrite;
  AssignFile(WAVFile, WAVName);

  {$I-}
  Rewrite(WAVFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function ExtractEntry(var FSBFile: file; const FSBEntry: TFSBEntry;
  const SearchPattern: AnsiString): Boolean;
var
  Name            : AnsiString;
  WAVFile         : file;
  CodeBuf         : array of Byte;
  SampleBuf       : array of Smallint;
  Handle,
  PatPos          : Longint;
  BitsPerSample,
  BlockAlign      : Word;
  BytesPerSecond,
  UncompSize      : Longword;
begin
  PatPos := SearchMem(@SearchPattern[1], @FEVBuffer[0], Length(SearchPattern),
    Length(FEVBuffer));

  if PatPos < 6 then
    Name := ''
  else
  begin
    PatPos := SearchMemReverse(@FEVBuffer[PatPos - 5], @FEVBuffer[0], 1,
      PatPos - 6);

    if PatPos < 0 then
      Name := ''
    else
      Name := PChar(@FEVBuffer[PatPos + 1])
  end;

  if Name = '' then
  begin
    Name := FSBEntry.Name + #000;
    SetLength(Name, StrLen(@Name[1]));
  end;

  // load compressed data (if this fails, there's no chance of recovery)
  SetLength(CodeBuf, FSBEntry.DataSize);
  {$I-}
  BlockRead(FSBFile, CodeBuf[0], Length(CodeBuf));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
  begin
    WriteLn('    seek error');
    Exit
  end;

  // open output wave file
  if not OpenWAVFile(WAVFile, Name) then
  begin
    WriteLn('    error opening output');
    Exit
  end;

  // load and decompress ADPCM data
  Handle := GetIMAManager(FSBEntry.BlockSize + FSBEntry.NumChannels * 4, 0,
    FSBEntry.NumChannels, IMA_ADPCM_Format_Default);

  IMALoadCodes(Handle, CodeBuf[0], Length(CodeBuf));

  SetLength(SampleBuf, IMAGetDecompressedSize(Handle));
  IMADecompress(Handle, SampleBuf[0]);

  // wave parameters
  UncompSize := Length(SampleBuf) * 2;
  BytesPerSecond := FSBEntry.Frequency * FSBEntry.NumChannels * 2;
  BlockAlign := FSBEntry.NumChannels * 2;
  BitsPerSample := 16;

  // prepare wave header
  Inc(UncompSize, 32);

  Move(UncompSize, WAVHeader[4], 4);
  Move(FSBEntry.NumChannels, WAVHeader[22], 2);
  Move(FSBEntry.Frequency, WAVHeader[24], 4);
  Move(BytesPerSecond, WAVHeader[28], 4);
  Move(BlockAlign, WAVHeader[32], 2);
  Move(BitsPerSample, WAVHeader[34], 2);

  Dec(UncompSize, 32);

  Move(UncompSize, WAVHeader[40], 4);

  // output wave data
  {$I-}
  BlockWrite(WAVFile, WAVHeader[0], Length(WAVHeader));
  BlockWrite(WAVFile, SampleBuf[0], Length(SampleBuf) * 2);
  {$I+}
  if IOResult <> 0 then
    WriteLn('    write error');

  IMAUnloadCodes(Handle);

  ReleaseIMAManager(Handle);

  SetLength(CodeBuf, 0);
  SetLength(SampleBuf, 0);

  CloseFile(WAVFile)
end;

procedure ExtractFSB(var FSBFile: file; FSBName: AnsiString;
  const FSBHeader: TFSBHeader);
var
  i         : Longint;
  FSBEntries: array of TFSBEntry;
  BytesRead : Longint;
  SearchPat : AnsiString;
begin
  FSBName := ExtractFileName(FSBName);
  FSBName := Copy(FSBName, 1,
    Length(FSBName) - Length(ExtractFileExt(FSBName)));

  for i := 0 to High(FSBList) do
    if AnsiLowerCase(FSBName) = AnsiLowerCase(FSBList[i]) then
      FSBName := FSBList[i];

  SetLength(FSBEntries, FSBHeader.NumEntries);

  // read directory
  for i := 0 to High(FSBEntries) do
    with FSBEntries[i] do
    begin
      BlockRead(FSBFile, EntrySize, SizeOf(EntrySize), BytesRead);
      if BytesRead <> SizeOf(EntrySize) then
      begin
        WriteLn('  read error');
        SetLength(FSBEntries, 0);
        Exit
      end;

      // only an EntrySize of 80 bytes is currently supported
      if EntrySize <> 80 then
      begin
        WriteLn('  WARNING: Unsupported entry encountered; no processing beyo' +
          'nd this point.');

        SetLength(FSBEntries, i);
        Break
      end;

      // read rest of entry data
      BlockRead(FSBFile, Name, EntrySize - SizeOf(EntrySize), BytesRead);
      if BytesRead <> EntrySize - SizeOf(EntrySize) then
      begin
        WriteLn('  read error');
        SetLength(FSBEntries, 0);
        Exit
      end;
    end;

  // process data
  for i := 0 to High(FSBEntries) do
  begin
    SetLength(SearchPat, Length(FSBName) + 5);

    Move(FSBName[1], SearchPat[1], Length(FSBName));
    SearchPat[Length(FSBName) + 1] := #000;
    Move(i, SearchPat[Length(FSBName) + 2], 4);

    with FSBEntries[i] do
      if not ExtractEntry(FSBFile, FSBEntries[i], SearchPat) then
        Break;
  end;

  SetLength(FSBEntries, 0)
end;

procedure ProcessFSB(var FSBFile: file; const FSBName: AnsiString);
var
  FSBHeader: TFSBHeader;
  BytesRead: Longint;
begin
  BlockRead(FSBFile, FSBHeader, SizeOf(FSBHeader), BytesRead);
  if (BytesRead <> SizeOf(FSBHeader)) or (FSBHeader.ID <> FSBHeaderID) then
  begin
    WriteLn('  not a valid (or supported) FSB file');
    Exit
  end;

  with FSBHeader do
  begin
    if SizeOf(FSBHeader) + DirSize + DataSize <>
      Longword(FileSize(FSBFile)) then
      WriteLn('  WARNING: File size suspect.');

    // currently, we only know entries of 80 bytes in size
    if NumEntries * 80 <> DirSize then
      WriteLn('  WARNING: Directory size suspect.');
  end;

  ExtractFSB(FSBFile, FSBName, FSBHeader)
end;

procedure ScanFiles(FileParam, NameFileParam: AnsiString);
var
  FileName,
  Dir,
  FSBName  : AnsiString;
  FSBFile  : file;
  SearchRec: TSearchRec;
begin
  LoadFEVBuffer(NameFileParam);

  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      FSBName := Dir + FileName;

      WriteLn(FSBName);

      // if this file is accessible, analyze it
      if OpenFSBFile(FSBFile, FSBName) then
      begin
        ProcessFSB(FSBFile, FSBName);
        Close(FSBFile)
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  FindClose(SearchRec)
end;

begin
  if not LoadDLL(ExtractFilePath(ParamStr(0))) then
  begin
    WriteLn(Format('Cannot load library %s.', [DLLName]));
    Exit
  end;

  if not DLLCanUse then
  begin
    WriteLn(Format('Incompatible version of library %s.', [DLLName]));
    Exit
  end;

  if ParamCount = 1 then
    ScanFiles(ParamStr(1), '')
  else if ParamCount > 1 then
    ScanFiles(ParamStr(1), ParamStr(2))
  else
    Writeln('Incorrect number of parameters.');
end.
